#include "configuration.h"
#include <fcntl.h>
#include <stddef.h>
#include <stdio.h>

#define DEFAULT_LOGIN_SESSION_USER "root"

/**
 * @brief Object properties enumeration
 */
enum
{
    NO_PROP,
    DAEMON_LOGIN_SESSION_COMMAND,
    DAEMON_LOGIN_SESSION_USER,
    DAEMON_SESSION_COMMAND,
    N_PROPERTIES
};
static GParamSpec *conf_properties[N_PROPERTIES] = {0};

/**
 * @brief Object private structure
 */
struct _DaemonConfiguration
{
    BaseObject parent_object;
    gchar* login_session_command;
    gchar* login_session_user;
    gchar* session_command;
    gchar* string;
};

G_DEFINE_FINAL_TYPE(DaemonConfiguration, daemon_configuration, BASE_TYPE_OBJECT)

/**
 * @brief Dispose the object
 */
static void daemon_configuration_dispose(GObject* object)
{
    DaemonConfiguration * self = DAEMON_CONFIGURATION(object);
    g_debug("daemon_configuration_dispose: %s", base_object_to_string(self));
    if(self->login_session_command)
    {
        g_free(self->login_session_command);
    }
    if(self->login_session_user)
    {
        g_free(self->login_session_user);
    }
    if(self->session_command)
    {
        g_free(self->session_command);
    }
    if(self->string)
    {
        g_free(self->string);
    }
    self->string = NULL;
    G_OBJECT_CLASS(daemon_configuration_parent_class)->dispose(object);
}
/**
 * @brief Set the object property
 */
static void daemon_configuration_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    (void)object;
    (void)value;
    (void)pspec;
    // DaemonConfiguration * self = DAEMON_CONFIGURATION(object);
    switch (property_id)
    {
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

/**
 * @brief Get the object property
 */
static void daemon_configuration_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DaemonConfiguration * self = DAEMON_CONFIGURATION(object);
    switch (property_id)
    {
    case DAEMON_LOGIN_SESSION_COMMAND:
        g_value_set_string(value, self->login_session_command);
        break;
    case DAEMON_LOGIN_SESSION_USER:
        g_value_set_string(value, self->login_session_user);
        break;
    case DAEMON_SESSION_COMMAND:
        g_value_set_string(value, self->session_command);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

/**
 * @brief Get the string representation of the object
 */
static const gchar* daemon_configuration_to_string(BaseObject* object)
{
    g_debug("Call daemon_configuration_to_string");
    DaemonConfiguration* self = DAEMON_CONFIGURATION(object);
    return self->string;
}

/**
 * @brief Class initialization
 */
static void daemon_configuration_class_init(DaemonConfigurationClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    BaseObjectClass *base_class = BASE_OBJECT_CLASS(class);

    gobject_class->dispose = daemon_configuration_dispose;
    gobject_class->set_property = daemon_configuration_set_property;
    gobject_class->get_property = daemon_configuration_get_property;

    base_class->to_string = daemon_configuration_to_string;

    conf_properties[DAEMON_LOGIN_SESSION_COMMAND] = g_param_spec_string("login-session-command", NULL, "Login session command", NULL, G_PARAM_READABLE);
    conf_properties[DAEMON_LOGIN_SESSION_USER] = g_param_spec_string("login-session-user", NULL, "Login session user", NULL, G_PARAM_READABLE);
    conf_properties[DAEMON_SESSION_COMMAND] = g_param_spec_string("session-command", NULL, "Session command", NULL, G_PARAM_READABLE);
    g_object_class_install_properties (gobject_class, N_PROPERTIES, conf_properties);
}

/**
 * @brief Object initialization
 */
static void daemon_configuration_init(DaemonConfiguration *self)
{
    self->login_session_command = NULL;
    self->login_session_user = NULL;
    self->session_command = NULL;
    self->string = NULL;
}
/**
 * @brief Create a new DaemonConfiguration object
 * 
 * read and parse the config_file
 * The format of the config file is as follows:
 * - line starting with # are comments and should be ignored
 * - line starting with login_session_command= is the command to start the login session
 * - line starting with login_session_user= is the user to start the login session
 * - line starting with session_command= is the command to start the session
 * 
 * @param config_file The path to the configuration file
 * @return The DaemonConfiguration object
 */
DaemonConfiguration* daemon_configuration_new(const gchar* config_file)
{   
    FILE* file = fopen(config_file, "r");
    if(!file)
    {
        g_critical("Error opening the configuration file %s: %s", config_file, strerror(errno));
        return NULL;
    }
    /**
     * read file line by line
     */
    DaemonConfiguration* conf = g_object_new(DAEMON_TYPE_CONFIGURATION, NULL);
    char* line = NULL;
    ssize_t read;
    gchar* ptr = NULL;
    int conf_valid = 0;
    size_t len = 0;
    conf->string = g_strconcat("DaemonConfiguration from ",
                        config_file,
                        ":\n",
                        NULL);
    ptr = conf->string;
    while ((read = getline(&line, &len, file)) != -1)
    {
        if (line[0] == '#')
        {
            free(line);
            line = NULL;
            continue;
        }
        /**
         * strip newline at the end
         */
        if (line[read - 1] == '\n')
        {
            line[read - 1] = '\0';
        }
        /**
         * parse line using strok
         */
        char* value = NULL;
        char* key = strtok_r(line, "=", &value);
        if(!key || !value)
        {
            free(line);
            line = NULL;
            continue;
        }
        /**
         * strim key and value at the begin and the end
         */
        while(*key == ' ')
        {
            key++;
        }
        char* end = key + strlen(key) - 1;
        while(*end == ' ')
        {
            *end = '\0';
            end--;
        }
        while(*value == ' ')
        {
            value++;
        }
        end = value + strlen(value) - 1;
        while(*end == ' ')
        {
            *end = '\0';
            end--;
        }
        conf_valid = 1;
        if (strncmp(key, "login_session_command", 22) == 0)
        {
            conf->login_session_command = g_strdup(value);
        } 
        else if (strncmp(key, "login_session_user", 19) == 0)
        {
            conf->login_session_user = g_strdup(value);
        }
        else if (strncmp(key, "user_session_command", 21) == 0)
        {
            conf->session_command = g_strdup(value);
        }
        else
        {
            g_warning("Ignore unknown configuration: [%s]:[%s]", key, value);
            conf_valid = 0;
        }
        if(conf_valid)
        {
            conf->string = g_strconcat(ptr, " - ", key, ": ", value, "\n", NULL);
            g_free(ptr);
            ptr = conf->string;
        }
        free(line);
        line = NULL;
    }
    if(line)
    {
        free(line);
    }
    fclose(file);
    if(conf->login_session_user == NULL)
    {
        conf->login_session_user = g_strdup("root");
    }
    if(conf->login_session_command == NULL || conf->session_command == NULL)
    {
        g_critical("login_session_command or session_command is not set in the configuration file");
        g_object_unref(conf);
        return NULL;
    }
    return conf;
}