#include "manager.h"
#include "configuration.h"
#include <gio/gio.h>
#include <pwd.h>
#include <grp.h>
#include <shadow.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#define PROCESS_MOM_TO 50 // in ms

#define DBUS_SERVER_NAME "dev.iohub.diya.SessionManager"
#define DBUS_SERVER_PATH "/dev/iohub/diya/SessionManager"
#define DBUS_SERVER_ERROR_NAME "dev.iohub.diya.SessionManager.Error"

typedef enum
{
    SESSION_MANAGER_STATE_IDLE,
    SESSION_MANAGER_STATE_LOGIN,
    SESSION_MANAGER_STATE_SESSION
} session_manage_state_t;

/**
 * @brief Object properties enumeration
 */
enum
{
    SESSION_MANAGER_NO_PROP,
    SESSION_MANAGER_CONFIGURATION,
    N_PROPERTIES
};

static GParamSpec *manager_properties[N_PROPERTIES] = {0};

/**
 * @brief Object private structure
 */
struct _SessionManager
{
    BaseObject parent_object;
    DaemonConfiguration *configuration;
    pid_t session_pid;
    pid_t login_session_pid;
    GMainLoop *loop;
    GDBusNodeInfo *introspection;
    guint bus_id;
    session_manage_state_t state;
    gchar* current_session_user;
};

G_DEFINE_FINAL_TYPE(SessionManager, session_manager, BASE_TYPE_OBJECT);

/**
 * @brief Dispose the object
 */
static void session_manager_dispose(GObject *object)
{
    SessionManager *self = SESSION_MANAGER(object);
    g_debug("session_manager_dispose: %s", base_object_to_string(self));
    if (self->configuration)
    {
        g_object_unref(self->configuration);
    }
    if (self->loop)
    {
        g_main_loop_unref(self->loop);
    }
    if (self->introspection)
    {
        g_dbus_node_info_unref(self->introspection);
    }
    if (self->bus_id > 0)
    {
        g_bus_unown_name(self->bus_id);
    }
    if (self->current_session_user)
    {
        g_free(self->current_session_user);
    }
    G_OBJECT_CLASS(session_manager_parent_class)->dispose(object);
}

/**
 * @brief Get the string representation of the object
 */
static const gchar *session_manager_to_string(BaseObject *object)
{
    (void)object;
    // SessionManager* self = SESSION_MANAGER(object);
    return "SessionManager";
}

/**
 * @brief Get the object property
 */
static void session_manager_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    SessionManager *self = SESSION_MANAGER(object);
    switch (property_id)
    {
    case SESSION_MANAGER_CONFIGURATION:
        g_value_set_pointer(value, self->configuration);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
        break;
    }
}

/**
 * @brief Class initialization
 */
static void session_manager_class_init(SessionManagerClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    BaseObjectClass *base_class = BASE_OBJECT_CLASS(class);
    gobject_class->dispose = session_manager_dispose;
    gobject_class->get_property = session_manager_get_property;

    base_class->to_string = session_manager_to_string;

    manager_properties[SESSION_MANAGER_CONFIGURATION] = g_param_spec_pointer("configuration", NULL, "Configuration", G_PARAM_READABLE);
    g_object_class_install_properties(gobject_class, N_PROPERTIES, manager_properties);
}

/**
 * @brief Object initialization
 */
static void session_manager_init(SessionManager *self)
{
    self->configuration = NULL;
    self->session_pid = -1;
    self->login_session_pid = -1;
    self->loop = g_main_loop_new(NULL, FALSE);
    self->introspection = NULL;
    self->bus_id = 0;
    self->state = SESSION_MANAGER_STATE_IDLE;
    self->current_session_user = NULL;
}

static gboolean session_user_auth(const char *username, const char *password)
{
    char *encrypted;
    struct passwd *pwd;
    struct spwd *spwd;
    /* Look up password and shadow password records for username */
    pwd = getpwnam(username);
    if (pwd == NULL)
    {
        g_critical("getpwnam: %s", strerror(errno));
        return FALSE;
    }
    spwd = getspnam(username);
    /*
    if (spwd == NULL)
    {
        g_critical("getspnam: %s", strerror(errno));
        return FALSE;
    }
    */

    if (spwd != NULL) /* If there is a shadow password record */
    {
        pwd->pw_passwd = spwd->sp_pwdp; /* Use the shadow password */
    }
    /* Encrypt password and erase cleartext version immediately */
    encrypted = crypt(password, pwd->pw_passwd);
    if (encrypted == NULL)
    {
        g_critical("crypt: %s", strerror(errno));
        return FALSE;
    }
    if (strcmp(encrypted, pwd->pw_passwd) == 0)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

static pid_t run_command_as(const gchar* command, const gchar* username)
{
    g_debug("run_command_as: start user session command [%s] for user %s", command, username);
    int numgroups;
    int numgroups_max = sysconf(_SC_NGROUPS_MAX) + 1;
    gid_t g_list[numgroups_max];
    struct passwd *pwd = getpwnam(username);
    if (pwd == NULL)
    {
        g_critical("run_command_as: Unable to call getpwnam: %s", strerror(errno));
        return -1;
    }
    if(initgroups(username, 0) != 0)
    {
        g_critical("run_command_as: Unable to call initgroups: %s", strerror(errno));
        return -1;
    }
    if (getgrouplist(username, pwd->pw_gid, g_list, &numgroups) == -1)
    {
        g_critical("run_command_as: Unable to call getgrouplist: %s", strerror(errno));
        return -1;
    }
    pid_t pid = fork();
    if (pid == 0)
    {
        if (setgid(pwd->pw_gid) == -1)
        {
            g_critical("run_command_as: Unable to call setgid: %s", strerror(errno));
            exit(1);
        }
        /*
        g_debug("run_command_as: setgroups: %d", numgroups);
        for(int i = 0; i < numgroups; i++)
        {
            g_debug("run_command_as: group: %d \n", g_list[i]);
        }
        */
        if (setgroups(numgroups, g_list) == -1)
        {
            g_critical("run_command_as: Unable to call setgroups: %s", strerror(errno));
            exit(1);
        }
        if (setuid(pwd->pw_uid) == -1)
        {
            g_critical("run_command_as: Unable to call setuid: %s", strerror(errno));
            exit(1);
        }
        char env_user[128];
        char* args[] = {"/bin/sh", "-c", (char*)command, NULL};
        char* env[1] = { NULL};
        g_snprintf(env_user, sizeof(env_user), "USER=%s", username);
        env[0] = env_user;
        execve(args[0], args, env);
        g_critical("execve: %s", strerror(errno));
        exit(-1);
    }
    return pid;
}

/**
 * @brief Start the user session
 */
static void start_user_session(SessionManager *self)
{
    if(self->current_session_user == NULL)
    {
        g_critical("start_user_session: Current session user is not set. No user logged in");
        return;
    }
    /**
     * wait for the login session to exit
     */
    if (self->login_session_pid > 0)
    {
        int status;
        pid_t pid = waitpid(self->login_session_pid, &status, 0);
        if(pid <= 0)
        {
            g_warning("start_user_session: login session is still running or waitpid error: %s. Force kill it", strerror(errno));
            // try to kill it
            (void)kill(self->login_session_pid, SIGKILL);
            pid = waitpid(self->login_session_pid, &status, 0);
        }
        g_debug("start_user_session: login session is exited: %d", pid);
        self->login_session_pid = 0;
    }
    if(self->session_pid > 0)
    {
        g_warning("start_user_session: Cannot start user session for user %s. A session is currently running", self->current_session_user);
        return;
    }
    gchar *session_command = NULL;
    g_object_get(self->configuration, "session-command", &session_command, NULL);
    if (session_command == NULL)
    {
        g_free(session_command);
        g_critical("start_user_session: session_command is not set in the configuration file, do nothing");
        return;
    }
    self->session_pid = run_command_as(session_command, self->current_session_user);
    g_free(session_command);
}

/**
 * @brief Method call handler on dbus interface
 */
static void session_manager_method_call(GDBusConnection *connection, const gchar *sender, const gchar *object_path, const gchar *interface_name, const gchar *method_name, GVariant *parameters, GDBusMethodInvocation *invocation, gpointer user_data)
{
    (void)connection;
    (void)sender;
    (void)object_path;
    (void)interface_name;
    (void)user_data;
    SessionManager *self = SESSION_MANAGER(user_data);
    g_debug("session_manager_method_call: %s, object path: %s, interface name: %s", method_name, object_path, interface_name);
    if (g_strcmp0(interface_name, DBUS_SERVER_NAME) != 0)
    {
        g_critical("Interface name mismatch: got %s, but expected %s", interface_name, DBUS_SERVER_NAME);
        g_dbus_method_invocation_return_dbus_error(invocation, DBUS_SERVER_ERROR_NAME, "Method not found");
        return;
    }
    if (g_strcmp0(method_name, "login") == 0)
    {
        const gchar *username;
        const gchar *password;
        g_variant_get(parameters, "(&s&s)", &username, &password);
        g_debug("Login request for user: %s", username);

        if (self->session_pid > 0)
        {
            g_warning("Cannot process login request for user %s. A session is currently running", username);
            g_dbus_method_invocation_return_dbus_error(invocation, DBUS_SERVER_ERROR_NAME, "A session is currently running");
            return;
        }
        if(session_user_auth(username, password))
        {
            g_dbus_method_invocation_return_value(invocation, g_variant_new("(b)", TRUE));
            /**
             * wait for login session to exit and start user session
             */
            if(self->current_session_user)
            {
                g_free(self->current_session_user);
                self->current_session_user = NULL;
            }
            self->current_session_user = g_strdup(username);
            self->state = SESSION_MANAGER_STATE_SESSION;
        }
        else
        {
            g_dbus_method_invocation_return_value(invocation, g_variant_new("(b)", FALSE));
        }
        // g_variant_unref(result);
    }
    else
    {
        g_critical("Method not found: %s", method_name);
        g_dbus_method_invocation_return_dbus_error(invocation, DBUS_SERVER_ERROR_NAME, "Method not found");
    }
}

/**
 * @brief Bus listener callback
 */
static const GDBusInterfaceVTable interface_vtable =
    {
        .method_call = session_manager_method_call,
        .get_property = NULL,
        .set_property = NULL,
};

/**
 * @brief Bus acquired callback
 */
static void on_bus_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    SessionManager *self = SESSION_MANAGER(user_data);
    g_debug("on_bus_acquired: %s", name);
    GError *error = NULL;
    if (!g_dbus_connection_register_object(
            connection,
            DBUS_SERVER_PATH,
            self->introspection->interfaces[0],
            &interface_vtable,
            user_data,
            NULL,
            &error))
    {
        g_critical("Failed to register object: %s", error->message);
        g_error_free(error);
        session_manager_event_loop_stop(self);
    }
}

/**
 * @brief Name acquired callback
 */
static void on_name_acquired(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    (void)connection;
    (void)user_data;
    g_debug("on_name_acquired: %s", name);
}

/**
 * @brief Name lost callback
 */
static void on_name_lost(GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    (void)connection;
    g_critical("DBus session manager name lost, probably because of user not having access rights on the bus: %s", name);
    SessionManager *self = SESSION_MANAGER(user_data);
    session_manager_event_loop_stop(self);
}

/**
 * @brief Create a new SessionManager object
 */
SessionManager *session_manager_new(const gchar *config_file)
{
    SessionManager *manager = g_object_new(SESSION_TYPE_MANAGER, NULL);
    manager->configuration = daemon_configuration_new(config_file);
    if (!manager->configuration)
    {
        return NULL;
    }
    // load xml content from gresource
    GError *err = NULL;
    GBytes *bytes = g_resources_lookup_data("/dev/iohub/diya/SessionManager/resources/introspection.xml", 0, &err);
    if (err != NULL)
    {
        g_critical("Unable to load introspection resource: %s", err->message);
        g_error_free(err);
        g_object_unref(manager);
        return NULL;
    }
    const char *data = g_bytes_get_data(bytes, NULL);
    g_debug("introspection xml: \n %s", data);
    manager->introspection = g_dbus_node_info_new_for_xml(data, &err);
    g_bytes_unref(bytes);
    if (err != NULL)
    {
        g_critical("Unable to parse introspection xml: %s", err->message);
        g_error_free(err);
        g_object_unref(manager);
        return NULL;
    }
    manager->bus_id = g_bus_own_name(G_BUS_TYPE_SYSTEM,
                                     DBUS_SERVER_NAME,
                                     G_BUS_NAME_OWNER_FLAGS_NONE,
                                     on_bus_acquired,
                                     on_name_acquired,
                                     on_name_lost,
                                     manager,
                                     NULL);

    return manager;
}

static void start_login_session(SessionManager *self)
{
    if(self->session_pid > 0)
    {
        g_warning("start_login_session: Cannot start login session. A session is currently running");
        return;
    }
    if(self->login_session_pid > 0)
    {
        g_warning("start_login_session: login session is already running");
        return;
    }
    if(self->current_session_user)
    {
        g_free(self->current_session_user);
        self->current_session_user = NULL;
    }
    gchar *login_session_command = NULL;
    g_object_get(self->configuration, "login-session-command", &login_session_command, NULL);
    if (login_session_command == NULL)
    {
        g_critical("login_session_command is not set in the configuration file, do nothing");
        return;
    }

    // get the login login_session_user from configuration
    gchar *login_session_user = NULL;
    g_object_get(self->configuration, "login-session-user", &login_session_user, NULL);
    if (login_session_user == NULL)
    {
        g_critical("login_session_user is not set in the configuration file, start the login session as root");
        return;
    }
    
    self->login_session_pid = run_command_as(login_session_command, login_session_user);
    g_free(login_session_user);
    g_free(login_session_command);
}

/**
 * @brief Process monitor
 */
static gboolean processes_monitor(gpointer user_data)
{
    int status;
    pid_t pid;
    SessionManager *self = SESSION_MANAGER(user_data);

    switch (self->state)
    {
    case SESSION_MANAGER_STATE_IDLE:
        if(self->session_pid > 0)
        {
            
            pid = waitpid(self->session_pid, &status, WNOHANG);
            if(pid == -1)
            {
                g_critical("session waitpid error: %s", strerror(errno));
                // try to kill it
                (void)kill(self->session_pid, SIGKILL);
                (void)waitpid(self->session_pid, &status, 0);
                self->session_pid = -1;
            }
            if(pid == 0)
            {
                return TRUE;
            }
            self->session_pid = -1;
        }
        if(self->login_session_pid > 0)
        {
            pid = waitpid(self->login_session_pid, &status, WNOHANG);
            if(pid == -1)
            {
                g_critical("login session waitpid error: %s", strerror(errno));
                // try to kill it
                (void)kill(self->login_session_pid, SIGKILL);
                (void)waitpid(self->login_session_pid, &status, 0);
                self->login_session_pid = -1;
            }
            if(pid == 0)
            {
                return TRUE;
            }
            self->login_session_pid = -1;
        }
        self->state = SESSION_MANAGER_STATE_LOGIN;
        break;
    case SESSION_MANAGER_STATE_LOGIN:
        start_login_session(self);
        self->state = SESSION_MANAGER_STATE_IDLE;
        break;
    case SESSION_MANAGER_STATE_SESSION:
        start_user_session(self);
        self->state = SESSION_MANAGER_STATE_IDLE;
        break;
    default:
        break;
    }
    return TRUE;
}

/**
 * @brief Start the session manager event loop
 */
void session_manager_event_loop_start(SessionManager *self)
{
    if (g_main_loop_is_running(self->loop))
    {
        g_warning("Session manager already started. do nothing");
        return;
    }
    g_debug("Session manager event loop");

    g_timeout_add(PROCESS_MOM_TO, processes_monitor, self);

    g_main_loop_run(self->loop);
}

/**
 * @brief Stop the session manager event loop
 */
void session_manager_event_loop_stop(SessionManager *self)
{
    g_main_loop_quit(self->loop);
}